;	BC.ASM - Setup for C program execution on IBM PC. i
;	LARGE CASE .OBJ VERSION.

;	EQUs define fields.
IBM		EQU	21H	;IBM DOS INTERRUPT NUMBER
_CICODE		EQU	1
COCODE		EQU	2
DIRECT_CODE	EQU	6
DIRECT_INPUT	EQU	7
VERSION		EQU	30H
READ		EQU	3FH
SETBLK		EQU	4AH
TERMINATE	EQU	4CH

BS		EQU	8
TAB		EQU	9
LF		EQU	0AH
CR		EQU	0DH
CONTS		EQU	19
CONTX		EQU	24

DEFAULT_SP	EQU	8192		;stack size. can be changed here or
					;with link /s option.
DGROUP	GROUP	DATA,UDATA


	EXTRN	MAIN:FAR

_PROG	segment byte public 'PROG'
	ASSUME	DS:DGROUP,CS:_PROG

	PUBLIC	EXIT,PUTCHAR,GETCHAR,PUTS

;	this is the actual entry point for a C88 program.
;	the following initilization must take place before a program
;	is actually called.

;	1)	calculate argc and argv.
;	2)	set DS to users data area.
;	3)	call main(argc,argv).

;	if program returns then goto exit.



	PUBLIC	_CSETUP
_CSETUP	PROC	FAR
;	The setup code is here.
;	DS:0 is program segment prefix.
;	SS is program DS.


	MOV	AX,DGROUP
	MOV	ES,AX
	MOV	ES:_PCB,DS	;SAVE _PCB FOR EXIT CALL

;	REMEMBER VERSION NUMBER

	MOV	AH,VERSION	;CODE TO ASK FOR VERSION
	INT	IBM
	CMP	AL,2		;IS IT 2 OR ABOVE ?
	JB	OLDVER
	MOV	ES:_MSDOS2,1	;TRUE IF VERSION 2 OR ABOVE
OLDVER:

;	CLEAR RESERVED MEMORY. EVERYTHING FROM UDATA TO SS:0 IS CLEARED.

	PUSH	ES
	MOV	BX,ES
	MOV	DX,OFFSET DGROUP:UDATA
;	BX:DX IS FIRST BYTE TO CLEAR. NORMALIZE ADDRESS.
	MOV	DI,DX
	AND	DI,0FH
	MOV	CL,4
	SHR	DX,CL
	ADD	BX,DX
	MOV	DX,DI
	MOV	SI,SS
	XOR	AX,AX
	CLD
CLR_LP:	CMP	BX,SI		;DONE WHEN AT SS:n
	JZ	CLR_DONE
	MOV	ES,BX		;SEGMENT TO CLEAR
	MOV	DI,DX		;OFFSET TO CLEAR
	MOV	CX,8		;CLEAR 18 BYTES AT A TIME
REP	STOSW
	INC	BX		;DO NEXT SEGMENT
	JMP	CLR_LP
CLR_DONE:
	POP	ES


;	CHOP MEMORY ALLOCATION FOR BOTH MALLOC AND EXEC

	PUSH	ES
	MOV	CL,4
	MOV	BX,SS
	MOV	AX,SP		;WANT FIRST PARA OVER STACK
	SHR	AX,CL
	ADD	BX,AX
	INC	BX		;PARAS NEEDED
	SUB	BX,ES:_PCB	;PARAS NEED TO KEEP
	MOV	ES,ES:_PCB
	MOV	AH,SETBLK	;REDUCE MEMORY ALLOCATED
	INT	IBM
	POP	ES

;	CALCULATE ARGC AND ARGV. THE COMMAND TAIL IS BETWEEN 129 AND
;	255. COMMNAD TAIL IN COPIED TO TOP OF STACK AND ARGV VECTOR
;	IS PLACED UNDER IT. THE COMMAND NAME IS LAMENTABLY NULL.

	MOV	SI,255
	MOV	BYTE PTR [SI],CR	;DONT FALL OFF END IF INVALID
	MOV	SI,128
	MOV	BYTE PTR [SI],' '	;DONT FALL OFF START EITHER (REPLACES LENGTH)
	MOV	CX,1		;NUMBER OF ARGUMENTS. ALWAYS (NULL) COMMAND
	MOV	SI,-1		;LENGTH OF ARGUMENTS.
;	FIND SIZE OF TAIL
CMDCOUNT:
	INC	SI
	MOV	AL,[SI+129]	;NEXT COMMAND CHAR
	CMP	AL,CR		;CR OR LF IS END OF LINE
	JZ	ENDCOUNT
	CMP	AL,LF
	JNZ	CMDCOUNT
ENDCOUNT:
	MOV	AX,0		;NEED ZERO AT END
	PUSH	AX
	TEST	SI,1		;NEED WORD ALLIGNED STACK
	JZ	WORD_A
	INC	SP		;ODD NOW SO EVEN LATER
WORD_A:	MOV	DI,SP		;REMEMBER LOC OF ZERO
	SUB	SP,SI		;MAKE ROOM FOR TAIL
	MOV	BP,SP		;BP CAN ADDRESS NEW TAIL
NTAIL:	DEC	SI
	CMP	SI,-1		;-1 WHEN DONE
	JZ	SAVE_ARGS
;	A NON WHITE SPACE PRECEEDED BY WHITE MEANS ANOTHER ARGUMENT
	MOV	AL,[SI+129]
	CMP	AL,' '		;SEE IF WHITE
	JZ	WHITE
	CMP	AL,TAB
	JNZ	NOT_WHITE
WHITE:	MOV	AL,0		;SET WHITE TO NULL FOR END OF STRING
	JMP	STUFF

;	IS PRECEEDING A WHITE?
NOT_WHITE:
	CMP	BYTE PTR [SI+128],' '
	JZ	NEW_ARG
	CMP	BYTE PTR [SI+128],TAB
	JNZ	STUFF
NEW_ARG:LEA	BX,[BP+SI]	;STORE ARGV ELEMENT
	PUSH	SS
	PUSH	BX
	INC	CX		;INCREMENT ARGC
STUFF:	MOV	[BP+SI],AL	;PUT TAIL CHAR ON STACK
	JMP	NTAIL

;	SAVE A POINTER TO A ZERO FOR ARGV[0]
SAVE_ARGS:
	PUSH	SS
	PUSH	DI

;	SAVE	ARGV AND ARGC
	MOV	BP,SP		;DONT PUSH SP AS 286 IS DIFFERENT
	PUSH	SS
	PUSH	BP		;*ARGV[]
	PUSH	CX		;ARGC

;	SET DS TO CORRECT VALUE

	MOV	AX,ES
	MOV	DS,AX



;	NOW READY FOR MAIN

	CALL	MAIN		;AND DO THE PROGRAM

;	FALL INTO EXIT IF RETURN

	MOV	AL,0		;SET RETURN CODE TO ZERO
	JMP	DOEXIT
_CSETUP	ENDP

;	exit();

EXIT	PROC	FAR
	POP	AX		;RETRIEVE RETURN CODE
	POP	AX
	POP	AX
DOEXIT:				;IN DOS 2.0 USE CODE 4C TO TERMINATE
				;AS THIS ALLOWS AN ERROR CODE
	CMP	_MSDOS2,0	;DOS 1.0 ?
	JZ	OLDEND
	MOV	AH,TERMINATE	;DOC 2.0 TERMINATE CODE
	INT	IBM


OLDEND:	PUSH	_PCB		;CREATE A LONG RETURN TO _PCB:0
	MOV	AX,0
	PUSH	AX
	RET
EXIT	ENDP


;	charactor = getchar();

GETCHAR	PROC	FAR
	CMP	LASTCH,0	;SAVED CHAR
	JZ	DOGC
GLAST:	MOV	AL,LASTCH	;RETURN LAST CHAR
	MOV	LASTCH,0
	JMP	GOTIN
DOGC:	MOV	AH,_CICODE	;CODE FOR ECHOED INPUT
	INT	IBM		;DO THE CONSOLE READ
GOTIN:	MOV	AH,0		;RETURN A WORD
	CMP	AL,26		;CONTROL z ?
	JZ	READNFG
	RET			;CHARACTER IS IN AL
READNFG:MOV	AX,-1		;-1 IF EOF
	RET
GETCHAR ENDP

;	putchar(character);

PUTCHAR	PROC	FAR
	POP	AX		;RETURN ADDRESS
	POP	BX
	POP	DX		;DL IS CHARACTER
	PUSH	DX
	PUSH	BX
	PUSH	AX
	CMP	DL,LF		;MUST CHANGE LF TO CR,LF
	JNZ	XCO
	MOV	DL,CR		;PRINT THE CR FIRST
	MOV	AH,COCODE
	INT	IBM		;DO THE WRITE
	MOV	DL,LF		;PUT THE LF BACK
XCO:	MOV	AH,COCODE
	INT	IBM		;DO THE WRITE
	RET
PUTCHAR ENDP

;	puts(string address);

PUTS	PROC	FAR
	POP	AX		;RETURN ADDRESS
	POP	CX
	POP	BX		;STRING ADDRESS
	POP	ES
	PUSH	ES
	PUSH	BX
	PUSH	CX
	PUSH	AX
PS_LP:	MOV	AL,ES:[BX]		;NEXT CHAR
	OR	AL,AL
	JZ	PS_END		;ZERO AT END OF STRING
	PUSH	ES
	PUSH	BX
	PUSH	AX		;CHAR
	CALL	PUTCHAR		;LET PUTCHAR TURN LF INTO CR,LF
	ADD	SP,2
	POP	BX
	POP	ES
	INC	BX		;STRING POINTER
	JMP	PS_LP
PS_END:	RET
PUTS	ENDP



	PUBLIC	CI, CO, CSTS

;	charactor = ci();

CI	PROC	FAR
	CMP	LASTCH,0	;SAVED CHAR
	JZ	DOCI
	MOV	AL,LASTCH	;RETURN LAST CHAR
	MOV	LASTCH,0
	MOV	AH,0
	RET
DOCI:	MOV	AH,DIRECT_INPUT	;CODE FOR DIRECT CONSOLE INPUT
	INT	IBM
	MOV	AH,0		;MAKE AN INT
	RET
CI	ENDP

;	charactor or zero = csts();

CSTS	PROC	FAR
	MOV	AH,DIRECT_CODE	;CODE FOR DIRECT CONSOLE IO
	MOV	DL,0FFH		;WANT A CHAR
	INT	IBM
	MOV	LASTCH,AL	;SAVE CHARACTER
	MOV	AH,0		;MAKE AN INT
	RET
CSTS	ENDP

;	co(character);

CO	PROC	FAR
	POP	AX		;RETURN ADDRESS
	POP	BX
	POP	DX
	PUSH	DX
	PUSH	BX
	PUSH	AX
	MOV	AH,DIRECT_CODE	;WANT A DIRECT CONSOLE OUTPUT
	INT	IBM
	RET
CO	ENDP

;	MEMORY MANAGEMENT FUNCTIONS: _MEMORY, _SHOWSP, _SETSP

	PUBLIC	_SETSP,_SHOWSP

_SHOWSP	PROC	FAR
	MOV	AX,SP
	ADD	AX,4
	RET
_SHOWSP	ENDP

_SETSP	PROC	FAR
	POP	DX		;CHOP THE STACK - DANGEROUS OPERATION
	POP	CX
	POP	BX		;NEW SP VALUE
	CMP	BP,SP		;ANY LOCALS
	JNZ	NOC
	MOV	BP,BX		;MUST MOVE THE BP TOO
NOC:	MOV	SP,BX
	PUSH	BX
	PUSH	CX
	PUSH	DX
	RET
_SETSP	ENDP

;	SEGMENT REGISTER FUNCTIONS: _SHOWDS, _SHOWCS, _SETDS

	PUBLIC	_SHOWDS,_SHOWCS,_SETDS

_SHOWDS	PROC	FAR
	MOV	AX,DS		;DS (AND SS) VALUE
	RET
_SHOWDS	ENDP

_SHOWCS	PROC	FAR
	MOV	AX,CS		;CS VALUE
	RET
_SHOWCS	ENDP

_SETDS	PROC	FAR
	POP	DX		;RETURN VALUE
	POP	BX
	POP	DS		;NEW DS
	PUSH	DS
	PUSH	BX
	PUSH	DX
	RET
_SETDS	ENDP




;	OS CALL USED BY OPEN, CREAT ETC.

;	VALUE=_OS(CODE,ARGUMENT);

	PUBLIC	_OS
_OS	PROC	FAR
	PUSH	BP
	MOV	BP,SP
	PUSH	DS
	MOV	AH,[BP+6]	;IBM CODE
	MOV	DX,[BP+8]	;ARGUMENT POINTER
	MOV	DS,[BP+10]	;DS:DX IS POINTER
	INT	IBM
	MOV	AH,0		;MAKE RETURN AN INTEGER
	POP	DS
	POP	BP
	RET
_OS	ENDP

_PROG	ENDS


;	ESEG IS THE SEGMENT FOR INITILIZED NON 'STATIC' ARRAYS AND
;	STRUCTURES. IT SHOULD BE BETWEEN CODE AND DATA.

ESEG	SEGMENT WORD PUBLIC 'ESEG'
ESEG	ENDS

;	DATA MUST BE LAST

DATA	SEGMENT WORD PUBLIC 'DATA'

	PUBLIC	_PCB,_MSDOS2,ERRNO	;USED BY CHAIN
_PCB	DW	0			;_PCB SEGMENT REGISTER
					;_PCB
_MSDOS2 DB	0			;1 IF VERSION 2.0 OR ABOVE
ERRNO	DW	0			;ERROR NUMBER FOR FLOATING POINT
LASTCH	DB	0
DATA	ENDS

UDATA	SEGMENT WORD PUBLIC 'DATA'
UDATA	ENDS

; NOTE A 8192 STACK IS DEFAULT.

STACK	SEGMENT STACK
STACK_FROM	LABEL BYTE
	DB	DEFAULT_SP DUP (?)
STACK	ENDS

	END	_CSETUP
